local cpml = require("cpml")

local ROT_Z = 0
local SCALE = 0.5
local W, H = 1024, 1024

local shader_shadow = love.graphics.newShader("shader_shadow.glsl")
local shader_main = love.graphics.newShader("shader_default.glsl")

local cube = love.graphics.newMesh({
  {"VertexPosition", "float", 3},
  -- {"VertexTexCoord", "float", 2},
}, {
  {-1,1,-1}, {1,1,1}, {1,1,-1}, {1,1,1}, {-1,-1,1}, {1,-1,1}, {-1,1,1}, {-1,-1,-1},
  {-1,-1,1}, {1,-1,-1}, {-1,-1,1}, {-1,-1,-1}, {1,1,-1}, {1,-1,1}, {1,-1,-1},
  {-1,1,-1}, {1,-1,-1}, {-1,-1,-1}, {-1,1,-1}, {-1,1,1}, {1,1,1}, {1,1,1}, {-1,1,1},
  {-1,-1,1}, {-1,1,1}, {-1,1,-1}, {-1,-1,-1}, {1,-1,-1}, {1,-1,1}, {-1,-1,1},
  {1,1,-1}, {1,1,1}, {1,-1,1}, {-1,1,-1}, {1,1,-1}, {1,-1,-1},
}, "triangles");

function getCamMatrix(kind, aspect, x, y, z, rotx, rotz)
  local proj
  if kind == "ortho" then
    proj = cpml.mat4.from_ortho(-10, 10, 10/aspect, -10/aspect, 0, 60)
  else
    proj = cpml.mat4.from_perspective(60, aspect, 0.01, 100)
  end

  local view = cpml.mat4.identity()
  view:resetTranslation(x, y, z)
  view:rotateZ(view, rotz)
  view:rotateX(view, -rotx)
  view:translateXYZ(view, 0, 0, -16)

  return view, proj
end

function makeCanvasOptions(w, h)
  local canvas = love.graphics.newCanvas(w, h)
  local depthcanvas = love.graphics.newCanvas(w, h, { format="depth24", readable=true })
  return { canvas, depth = true, depthstencil = depthcanvas }
end

local shadow_canvas_options = makeCanvasOptions(W, H)
local main_canvas_options = makeCanvasOptions(W, H)

function drawBegin(shader, view, proj, canvasOptions)
  -- local view, proj = self.cam:getMatrices(self.projectionType, self.w/self.h)
  shader:send("view", "column", view)
  shader:send("proj", "column", proj)

  love.graphics.push()
  love.graphics.setMeshCullMode("front")
  love.graphics.setDepthMode("lequal", true)
  love.graphics.setCanvas(canvasOptions)
  love.graphics.clear(0.3, 0.3, 0.3)

  love.graphics.setShader(shader)
  love.graphics.setColor(1,1,1)
end

function drawEnd()
  love.graphics.pop()
  love.graphics.setColor(1,1,1)
  love.graphics.setShader()
  love.graphics.setCanvas()
end

function drawAllItems(sh)
  -- column
  love.graphics.setColor(0, .3, .6)
  sh:send('model', "column", {
    0.5, 0, 0, 0,
    0, .5, 0, 0,
    0, 0, 10, 0,
    0, 0, 0, 1,
  })
  love.graphics.draw(cube)

  -- small cube
  love.graphics.setColor(0, .6, .3)
  sh:send('model', "column", {
    .5, 0, 0, 0,
    0, .5, 0, 0,
    0, 0, .5, 0,
    2, 5, 3, 1,
  })
  love.graphics.draw(cube)

  -- floor
  love.graphics.setColor(0, .6, .6)
  sh:send('model', "column", {
    6, 0, 0, 0,
    0, 6, 0, 0,
    0, 0, .1, 0,
    0, 0, 0, 1,
  })
  love.graphics.draw(cube)
end

function love.update()
  if love.keyboard.isDown("q") then ROT_Z = ROT_Z - 0.02 end
  if love.keyboard.isDown("e") then ROT_Z = ROT_Z + 0.02 end
  if love.keyboard.isDown("space") then ROT_Z = 0 end
  if love.keyboard.isDown("x") then ROT_Z = 0.75 end
end

function love.draw()
  local shadow_view, shadow_proj
  do -- shadow map
    shadow_view, shadow_proj = getCamMatrix("ortho", W/H, 0, 0, 0, 5, 0.75)

    drawBegin(shader_shadow, shadow_view, shadow_proj, shadow_canvas_options)
      drawAllItems(shader_shadow)
    drawEnd()

    love.graphics.draw(shadow_canvas_options.depthstencil, 0, 20, 0, SCALE) -- to screen
    love.graphics.print("shadow map", 5, 25)
  end


  do -- main (perspective) cam
    local main_view, main_proj = getCamMatrix("perspective", W/H, 0, 0, 0, 5, ROT_Z)

    shader_main:send("debug", 0)
    shader_main:send("shadowMap", shadow_canvas_options.depthstencil)
    shader_main:send("shadowView", shadow_view)
    shader_main:send("shadowProj", shadow_proj)

    drawBegin(shader_main, main_view, main_proj, main_canvas_options)
      drawAllItems(shader_main)
    drawEnd()

    love.graphics.draw(main_canvas_options[1], W * SCALE + 10, 20, 0, SCALE) -- to sceen
    love.graphics.print("main", W*SCALE+10+5, 25)

    -- closestDepth
    shader_main:send("debug", 1)
    drawBegin(shader_main, main_view, main_proj, main_canvas_options)
      drawAllItems(shader_main)
    drawEnd()
    love.graphics.draw(main_canvas_options[1], 1024 + 20, 20, 0, 0.25) -- to sceen
    love.graphics.print("closest", 1024 + 25, 25)

    -- currentDepth
    shader_main:send("debug", 2)
    drawBegin(shader_main, main_view, main_proj, main_canvas_options)
      drawAllItems(shader_main)
    drawEnd()
    love.graphics.draw(main_canvas_options[1], 1024 + 20, 256 + 30, 0, 0.25) -- to sceen
    love.graphics.print("current", 1024 + 25, 256 + 35)
  end

  love.graphics.print("ROT_Z: " .. ROT_Z, 2, 2)
end
